/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2008 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#ifndef _CSPInstanceInformationCallback_hh_
#define _CSPInstanceInformationCallback_hh_

/*
 * it is assumed that
 * - domains definitions are well formed (no duplicates)
 */

#include <set>
#include <map>

#include "ExpressionParser.hh"
#include "CSPParserCallback.hh"

namespace CSPXMLParser
{
  using namespace std;

  // definition of the functions which are called by the parser when it
  // reads new information in the file. These functions are in charge of
  // creating the data structure that the solver needs to do his job
  class CSPInstanceInformationCallback : public CSPParserCallback
  {
  public:
    unsigned int nbVar,nbConstr,maxArity,maxDomSize;
    unsigned int nbPredicateConstr,nbRelationConstr;
    map<string,int> globalConstraints; // name and number of
				       // occurrences of a given
				       // global constraint
  private:
    unsigned int curDomSize;

    string currentConstraintName,currentConstraintReference;
    CSPDefinitionType currentConstraintType;

  public:
    CSPInstanceInformationCallback()
    {
      nbVar=0;
      nbConstr=0;
      maxArity=0;
      maxDomSize=0;

      nbPredicateConstr=0;
      nbRelationConstr=0;
    }

    virtual void beginDomain(const string & name, int idDomain, int nbValue) 
    {
      curDomSize=0;
    }

    void addDomainValue(int v) 
    {
      curDomSize++;
    }

    virtual void addDomainValue(int first,int last) 
    {
      curDomSize+=last-first+1;
    }

    virtual void endDomain() 
    {
      maxDomSize=std::max(maxDomSize,curDomSize);
    }

    virtual void beginVariablesSection(int nbVariables) 
    {
      nbVar=0;
    }
  
    virtual void addVariable(const string & name, int idVar,
			     const string & domain, int idDomain) 
    {
      ++nbVar;
    }

    virtual void beginConstraint(const string & name, int idConstr,
				 int arity, 
				 const string & reference, 
				 CSPDefinitionType type, int id,
				 const ASTList &scope)
    {
      nbConstr++;
      maxArity=std::max(maxArity,(unsigned int)arity);

      switch(type)
      {
      case RelationType:
	nbRelationConstr++;
	break;

      case PredicateType:
	nbPredicateConstr++;
	break;

      case GlobalConstraintType:
	// get rid of "global:"
	if (reference.substr(0,7)=="global:")
	  globalConstraints[reference.substr(7)]++;
	else
	  globalConstraints[reference]++;
	break;
      }
    }
  };

} // namespace

#endif

