/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2008 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#ifndef _CSPVerifierCallback_hh_
#define _CSPVerifierCallback_hh_

#include <set>
#include <map>
#include <algorithm>

#include "ExpressionParser.hh"
#include "CSPParserCallback.hh"

#define ShowNumericalIDS

namespace CSPXMLParser
{
  using namespace std;
  
  class Instantiation: public VariableValuation
  {
  private:
    struct Value
    {
      bool isset; 
      FunctionValue val;
      string varName;

      Value() {isset=false;}
      Value(FunctionValue v) {isset=true;val=v;}
    };

    vector<Value> val;

  public:
    void clear()
    {
      val.clear();
    }

    int size() const
    {
      return val.size();
    }

    void setVarValue(int i, FunctionValue v)
    {
      if (i>=val.size())
	val.resize(i+1);

      val[i]=Value(v);
    }

    virtual FunctionValue getVarValue(unsigned int id) const
    {
      if (id<val.size() && val[id].isset)
	return val[id].val;
      else
	throw runtime_error("variable is not assigned");
    }

    void setVarName(unsigned int id, const string &name)
    {
      val[id].varName=name;
    }

    const string &getVarName(unsigned int id) const
    {
      return val[id].varName;
    }
  };

  class CheckedConstraint
  {
  public:
    static bool MAXCSPmode;
    static int nbViolatedConstraints;

    void dumpParameters(const AST &args,
			const VariableValuation &instantiation)
    {
      cout << "with\n";
      for(int i=0;i<args.size();++i)
      {
	cout << "  parm[" << i << "]=" 
	     << args[i].value(instantiation) 
	     << "   ";
	args[i].prefixExpression(cout);
	cout << endl;
      }
    } 

    /**
     * return true iff the constraint is satisfied
     */
    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)=0;
  };

  int CheckedConstraint::nbViolatedConstraints=0;
  bool CheckedConstraint::MAXCSPmode=false;

  class Domain
  {
  private:
    set<int> domain;
  public:
    void add(int val)
    {
      domain.insert(val);
    }

    void add(int first, int last)
    {
      for(int val=first;val<=last;++val)
	domain.insert(val);
    }

    bool belong(int val)
    {
      return domain.find(val)!=domain.end();
    }
  };

  class Predicate: public CheckedConstraint
  {
  private:
    AST *tree;

  public:
    Predicate()
    {
      tree=NULL;
    }

    void setAST(AST *tree)
    {
      this->tree=tree;
    }

    void dump(ostream &s)
    {
      tree->expression(s);
    }

    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      Instantiation parmsInstantiation;

      for(int i=0;i<args.size();++i)
	parmsInstantiation.setVarValue(i,args[i].value(instantiation));

      if (!tree->value(parmsInstantiation))
      {
	if (MAXCSPmode)
	  return false;

	cout << "ERROR: constraint " << constraintName 
	     << " is unsatisfied" << endl;

	cout << "Predicate is ";
	dump(cout);
	cout << endl;

	dumpParameters(args,instantiation);

	return false;
      }

      return true;
    }
  };

  class Relation: public CheckedConstraint
  {
  private:
    unsigned int arity;
    unsigned int expectedNbTuples;
    RelType type;
    vector<vector<int> > tuples;

    bool tuplesAreSorted; // are tuples sorted ?

  public:
    void init(int arity, int nbTuples, RelType relType) 
    {
      this->arity=arity;
      expectedNbTuples=nbTuples;
      type=relType;
    }

    void addTuple(unsigned int arity, int tuple[]) 
    {
      if (arity!=this->arity)
      {
	cout << "ERROR: a tuple has an unexpected arity" << endl;
	exit(1);
      }

      vector<int> tmptuple;

      for(unsigned int i=0;i<arity;++i)
	tmptuple.push_back(tuple[i]);

      if (!tuples.empty())
	tuplesAreSorted=tuplesAreSorted && (tuples.back()<tmptuple);

      tuples.push_back(tmptuple);
    }

    void sortTuples()
    {
      sort(tuples.begin(),tuples.end());
    }

    void endTuples()
    {
      if (!tuplesAreSorted)
      {
	sortTuples();
	tuplesAreSorted=true;
      }
    }

    void dumpTuples(ostream &s)
    {
      s << "Relation contains ";
      for(unsigned int i=0;i<tuples.size();++i)
      {
	vector<int> &tuple=tuples[i];

	s << "(" << tuple[0];
	for(unsigned int j=1;j<tuple.size();++j)
	  s << "," << tuple[j];
	s << ")";
      }
      s << endl;
    }

    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      if (args.size()!=arity)
      {
	cout << "ERROR: number of parameters differs from arity"
	     << " (relation check) in constraint"
	     << constraintName << endl;
	exit(1);
      }

      vector<int> parmTuple;
      for(unsigned int i=0;i<args.size();++i)
	parmTuple.push_back(args[i].value(instantiation));

      bool found=binary_search(tuples.begin(),tuples.end(),parmTuple);

      if (found)
      {
	if (type==REL_SUPPORT)
	  return true;

	if (type==REL_SOFT)
	{
	  throw runtime_error("unimplemented");
	}

	if (MAXCSPmode)
	  return false;

	cout << "ERROR: relation in constraint "
	     << constraintName << " is not satisfied"
	     << endl;

	dumpParameters(args,instantiation);

	if (type==REL_SUPPORT)
	  cout << "Relation contains support tuples" << endl;
	else
	  cout << "Relation contains conflict tuples" << endl;

	dumpTuples(cout);

	return false;
      }

      return true;
    }
  };

  class AllDifferent: public CheckedConstraint
  {
  public:
    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      set<int> S;

      if (args.size()==0)
      {
	cout << "ERROR: no variable in allDifferent constraint "
	     << constraintName << endl;
	exit(1);
      }

      const ASTList &parms=args[0].list();

      for(unsigned int i=0;i<parms.size();++i)
      {
	S.insert(parms[i].value(instantiation));
      }

      if (S.size()!=parms.size())
      {
	if (MAXCSPmode)
	  return false;

 	cout << "ERROR: some variables are equal in allDifferent constraint "
	     << constraintName << endl;

	dumpParameters(parms,instantiation);

	return false;
      }

      return true;
    }
  };


  class WeightedSum: public CheckedConstraint
  {
  public:
    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      const AST &sum=args[0];
      const AST &op=args[1];
      const AST &rhs=args[2];

      int LHS,RHS;
      bool verified=false;

      LHS=0;
      for(int i=0;i<sum.size();++i)
	LHS+=sum[i]["coef"].getInteger()*sum[i]["var"].value(instantiation);

      RHS=rhs.value(instantiation);

      switch(op.getType())
      {
      case SYMB_EQ:
	verified=(LHS==RHS);
	break;
      case SYMB_NE:
	verified=(LHS!=RHS);
	break;
      case SYMB_GE:
	verified=(LHS>=RHS);
	break;
      case SYMB_GT:
	verified=(LHS>RHS);
	break;
      case SYMB_LE:
	verified=(LHS<=RHS);
	break;
      case SYMB_LT:
	verified=(LHS<RHS);
	break;
      default:
	throw runtime_error("wrong operator in global:weightedSum constraint");
      }

      if (!verified)
      {
	if (MAXCSPmode)
	  return false;

 	cout << "ERROR: constraint global:weightedSum "
	     << constraintName << " is not verified "<< endl;

	//dumpParameters(args,instantiation);

	return false;
      }

      return true;
    }
  };

  class Element: public CheckedConstraint
  {
  public:
    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      const AST &index=args[0];
      const AST &table=args[1];
      const AST &value=args[2];

      bool verified=false;

      int INDEX; // INDEX in the array (starting from 0)

      INDEX=index.value(instantiation)-1;

      if (INDEX<0 || INDEX>=table.size())
	verified=false;
      else
	verified=(table[INDEX].value(instantiation)==
		  value.value(instantiation));

      if (!verified)
      {
	if (MAXCSPmode)
	  return false;

 	cout << "ERROR: constraint global:element "
	     << constraintName << " is not verified"<< endl;

	//dumpParameters(args,instantiation);

	return false;
      }

      return true;
    }
  };

  class Cumulative: public CheckedConstraint
  {
  private:
    struct Event
    {
      int time;
      int delta; // increase or decrease of ressource consumption

      Event(): time(0), delta(0) {}
      Event(int t, int d): time(t), delta(d) {}

      // chronological order (and in case of duce, ascending order on delta)
      bool operator <(const Event &e) const 
      {
	return time<e.time || (time==e.time && delta<e.delta);
      } 
    };
  public:
    virtual bool checkConstraint(const string &constraintName,
				 const ASTList &args,
				 const VariableValuation &instantiation)
    {
      vector<Event> eventList;
      ostringstream msg;

      const AST &tasks=args[0];
      const AST &limit=args[1];

      int LIMIT,ORIGIN,DURATION,END,HEIGHT;
      bool verified=true;

      for(int i=0;i<tasks.size();++i)
      {
	const AST &desc=tasks[i];
	int missingParm=0;

	if (desc.hasKey("origin"))
	{
	  ORIGIN=desc["origin"].value(instantiation);
	}
	else
	  ++missingParm;

	if (desc.hasKey("duration"))
	{
	  DURATION=desc["duration"].value(instantiation);
	}
	else
	  ++missingParm;

	if (desc.hasKey("end"))
	{
	  END=desc["end"].value(instantiation);
	}
	else
	  ++missingParm;

	if (missingParm>1)
	{
	  throw runtime_error("at least two of (ORIGIN,DURATION,END) must be specified in cumulative constraint");
	}
	else
	  if (missingParm==0)
	  {
	    if (ORIGIN+DURATION!=END)
	    {
	      msg << "duration is incoherent with origin and end";
	      verified=false;
	    }
	  }
	  else
	  {
	    if (!desc.hasKey("origin"))
	      ORIGIN=END-DURATION;

	    if (!desc.hasKey("end"))
	      END=ORIGIN+DURATION;

	    if (!desc.hasKey("duration"))
	      DURATION=END-ORIGIN;
	  }

	if (DURATION<0)
	{
	  msg << "negative duration";
	  verified=false;
	}

	if (!desc.hasKey("height"))
	  throw runtime_error("missing HEIGHT in cumulative constraint");

	HEIGHT=desc["height"].value(instantiation);

	eventList.push_back(Event(ORIGIN,+HEIGHT));
	eventList.push_back(Event(END,-HEIGHT));
      }

      sort(eventList.begin(),eventList.end());

      LIMIT=limit.value(instantiation);

      int current=0; // current ressource consumption

      for(int i=0;i<eventList.size();++i)
      {
	current+=eventList[i].delta;

	if (current>LIMIT)
	{
	  msg << "limit exceeded at time " << eventList[i].time;
	  verified=false;
	  break;
	}
      }

      if (!verified)
      {
	if (MAXCSPmode)
	  return false;

 	cout << "ERROR: constraint global:cumulative "
	     << constraintName << " is not verified "<< endl;
	cout << "  " << msg.str() << endl;

	//dumpParameters(args,instantiation);

	return false;
      }

      return true;
    }
  };

  // definition of the functions which are called by the parser when it
  // reads new information in the file. These functions are in charge of
  // creating the data structure that the solver needs to do his job
  class CSPVerifierCallback : public CSPParserCallback
  {
  private:
    Instantiation instantiation; // instantiation that must be checked
    vector<string> varNames;

    unsigned int nbVar,scopePos;

    CheckedConstraint *currentConstraint;

    Domain *currentDomain;
    vector<Domain *> domainList;

    Relation *currentRelation;
    vector<Relation *> relationList;

    Predicate *currentPredicate;
    vector<Predicate *> predicateList;

    string currentConstraintName,currentConstraintReference;
    CSPDefinitionType currentConstraintType;

    map<string,CheckedConstraint *> globalConstraints;

    int nbConstraints;
    bool listUnsatisfiedConstraints; // don't stop at the first
				     // falsified constraint

  public:
    CSPVerifierCallback()
    {
      listUnsatisfiedConstraints=false;

      nbConstraints=0;
      currentDomain=NULL;

      globalConstraints["global:alldifferent"]=new AllDifferent;
      globalConstraints["global:weightedsum"]=new WeightedSum;
      globalConstraints["global:element"]=new Element;
      globalConstraints["global:cumulative"]=new Cumulative;
    }

    void clearVarValues()
    {
      instantiation.clear();
    }

    void listAllUnsatisfiedConstraints()
    {
      listUnsatisfiedConstraints=true;
    }

    bool listMode() const
    {
      return listUnsatisfiedConstraints;
    }

    void setVarValue(unsigned int i, int val)
    {
      instantiation.setVarValue(i,val);
    }

    void setMAXCSPmode()
    {
      CheckedConstraint::MAXCSPmode=true;
    }

    bool getMAXCSPmode() const
    {
      return CheckedConstraint::MAXCSPmode;
    }

    int getNbViolatedConstraints()
    {
      return CheckedConstraint::nbViolatedConstraints;
    }

    int getNbSatisfiedConstraints()
    {
      return nbConstraints-CheckedConstraint::nbViolatedConstraints;
    }

    virtual void beginDomain(const string & name, int idDomain, int nbValue) 
    {
      currentDomain=new Domain;
      if (idDomain>=(int)domainList.size())
	domainList.resize(idDomain+1);

      domainList[idDomain]=currentDomain;
    }

    void addDomainValue(int v) 
    {
      currentDomain->add(v);
    }

    virtual void addDomainValue(int first,int last) 
    {
      currentDomain->add(first,last);      
    }

    virtual void beginVariablesSection(int nbVariables) 
    {
      nbVar=0;
    }
  
    virtual void addVariable(const string & name, int idVar,
			     const string & domain, int idDomain) 
    {
      ++nbVar;

      if (idVar>=(int)instantiation.size())
      {
	cout << "ERROR: variable " << name << " has no value!" << endl;
	exit(1);
      }

      if (!domainList[idDomain]->belong(instantiation.getVarValue(idVar)))
      {
	cout << "ERROR: variable " << name << " has value " 
	     << instantiation.getVarValue(idVar) 
	     << " which doesn't belong to its domain " << domain << endl;
	exit(1);
      }

      varNames.push_back(name);
    }

    virtual void endVariablesSection() 
    {
      if (nbVar!=instantiation.size())
      {
	cout << "ERROR: the number of variables in the model is different from the number of variable in the instance!" << endl;
	exit(1);
      }

#ifdef debug
      for(int i=0;i<instantiation.size();++i)
	cout << instantiation.getVarName(i) 
	     << "=" << instantiation.getVarValue(i) << endl;
#endif
    }

    virtual void beginRelation(const string & name, int idRel,
			       int arity, int nbTuples, RelType relType) 
    {
      currentRelation=new Relation;
      if (idRel>=(int)relationList.size())
	relationList.resize(idRel+1);

      relationList[idRel]=currentRelation;

      currentRelation->init(arity,nbTuples,relType);
    }

    virtual void addRelationTuple(int arity, int tuple[]) 
    {
      currentRelation->addTuple(arity,tuple);
    }

    virtual void endRelation() 
    {
      currentRelation->endTuples();
      currentRelation=NULL;
    }

    virtual void beginPredicate(const string &name, int idPred) 
    {
      currentPredicate=new Predicate;
      if (idPred>=(int)predicateList.size())
	predicateList.resize(idPred+1);

      predicateList[idPred]=currentPredicate;
    }

    virtual void addFormalParameter(int pos, const string &name, 
				    const string & type) 
    {
    }

    virtual void predicateExpression(AST *tree) 
    {
      currentPredicate->setAST(tree);
    }

    virtual void predicateExpression(const string &expr) 
    {
      // we don't use this one
      throw runtime_error("unimplemented");
    }

    virtual void endPredicate() 
    {
      currentPredicate=NULL;
    }

    virtual void beginConstraint(const string & name, int idConstr,
				 int arity, 
				 const string & reference, 
				 CSPDefinitionType type, int id,
				 const ASTList &scope)
    {
      nbConstraints++;

      currentConstraintName=name;
      currentConstraintReference=reference;
      currentConstraintType=type;

      switch(currentConstraintType)
      {
      case RelationType:
	currentConstraint=relationList[id];
	break;

      case PredicateType:
	currentConstraint=predicateList[id];
	break;

      case GlobalConstraintType:
	if (globalConstraints.find(reference)==
	    globalConstraints.end())
	{
	  cout << "ERROR: unknown (or unimplemented) global constraint " 
	       << reference << " in constraint " 
	       << currentConstraintName << endl;
	  exit(1);
	}

	currentConstraint=globalConstraints[reference];
	break;
      }
    }

    virtual void constraintParameters(const ASTList &args) 
    {
      if (!currentConstraint->checkConstraint(currentConstraintName,
					      args,instantiation))
      {
	if (CheckedConstraint::MAXCSPmode || listUnsatisfiedConstraints)
	  CheckedConstraint::nbViolatedConstraints++;
	else
	  exit(1);
      }
    }

    virtual void endConstraint() 
    {
    }


    /********************************************************************/
  };

} // namespace
#endif

